package algorithm;

import java.util.ArrayList;
import java.util.List;

import common.Numbers;
import common.TypeSafeMap;
import common.TypeSafeSet;

import base data.Graph;
import base data.Node;

/**
 * Role-based implementation of <a href="http://en.wikipedia.org/wiki/Dijkstra's_algorithm">Dijkstra's algorithm</a>
 * using OT/J. De-optimized version to closer mimic the Ruby example.
 */
public team class CalculateShortestPath {

	// general definition of the computation task:
	TraversedGraph geometries;
	TypeSafeSet<DistanceLabeledGraphNode> unvisited;
	DistanceLabeledGraphNode destination;
	Map map;

	// temporary data used while executing the algorithm
	Intersection currentIntersection;
	DistanceLabeledGraphNode current;
	Neighbor southNeighbor, eastNeighbor;
	
	// result data accumulated while executing the algorithm
	TypeSafeMap<data.Node,data.Node> pathTo;
	List<data.Node> path;
	
	/** Answer the computed shortest path from origin to destination. */
	public List<data.Node> getPath() {
		return this.path;
	}
	
	public CalculateShortestPath(Node as DistanceLabeledGraphNode originNode, 
			 					 Node as DistanceLabeledGraphNode targetNode,
			 					 Graph as TraversedGraph geometries,
			 					 Graph as Map map) 
	{
		// setup all nodes from geometries as unvisited:
		this.unvisited = new TypeSafeSet<DistanceLabeledGraphNode>();
		for(int i=0; i<geometries.getNumNodes(); i++) this.unvisited.add(geometries.getNodeAt(i));
		
		// use this to look up neighborship:
		this.geometries = geometries;

		// use the map to lookup distances:
		this.map = map;
		
		// initialize result structures:
		this.pathTo = new TypeSafeMap<data.Node, data.Node>();
		this.path = new ArrayList<data.Node>();
		
		// setup the origin node:
		originNode.tentativeDistance = 0;
		
		// that's where we want to travel to:
		this.destination = targetNode;
		
		// prepare for the first round:
		rebind (originNode, originNode);
	}

	void rebind(Node as Intersection currentIntersection, DistanceLabeledGraphNode currentWithDistance) {
		this.currentIntersection = currentIntersection;
		this.current = currentWithDistance;
		this.southNeighbor = geometries.southNeighborOf(currentIntersection);
		this.eastNeighbor = geometries.eastNeighborOf(currentIntersection);
	}

	/**
	 * Role TraversedGraph provides all contained nodes as {@link DistanceLabeledGraphNode}s. 
	 */
	protected team class TraversedGraph playedBy Graph {		
		int getNumNodes() -> int getNumNodes();
		DistanceLabeledGraphNode getNodeAt(int i) -> Node getNodeAt(int i);
		Neighbor eastNeighborOf(Intersection node) -> Node eastNeighborOf(Node node);
		Neighbor southNeighborOf(Intersection node) -> Node southNeighborOf(Node node);
	}
	
	/**
	 * Role Map looks up the distance between to nodes.
	 * Fortunately the Graph already provides this functionality, otherwise we would need to do more here.
	 */
	protected class Map playedBy Graph {

		int getDistance(DistanceLabeledGraphNode from, DistanceLabeledGraphNode to) -> int getDistance(Node from, Node to);
		
		// Ruby example has two more methods, which are, however unused.
	}

	/**
	 * View a node as an intersection which can answer all its (unvisited) neighbors.
	 */
	protected class Intersection playedBy Node {
		public List<Neighbor> unvisitedNeighbors() {
			List<Neighbor> retVal = new ArrayList<Neighbor>();
			if (southNeighbor != null && unvisited.safeContains(asDLGN(southNeighbor)))
				retVal.add(southNeighbor);
			if (eastNeighbor != null && unvisited.safeContains(asDLGN(eastNeighbor)))
				retVal.add(eastNeighbor);
			return retVal;
		}
	}

	/**
	 * Role DistanceLabeledGraphNode attaches a tentative distance-from-origin to a node
	 */
	protected class DistanceLabeledGraphNode playedBy Node {
		public int tentativeDistance = Integer.MAX_VALUE; // initially we assume the worst.
		
		String getName() -> String getName();
	}
	
	/** Given any node switch to its DistanceLabeledGraphNode role. */
	protected DistanceLabeledGraphNode asDLGN (Node as DistanceLabeledGraphNode node) { 
		return node; 
	}

	protected class Neighbor playedBy Node {
		/** 
		 * if newDistance is better than the current tentativeDistance use it from now on.
		 * @return true if the node has been relabled by this operation.
		 */
		public boolean relableAs(DistanceLabeledGraphNode self, int newDistance) {
			if (newDistance < self.tentativeDistance) {
				self.tentativeDistance = newDistance;
				return true;				
			}
			return false;
		}		
	}

	/**
	 * Execute Dijkstra's algorithm and return the node that serves as a handle to the computed path.
	 */
	public data.Node execute() {
		// The main loop

		// bottom part of the loop:
		for(Neighbor v : this.currentIntersection.unvisitedNeighbors()) {
			DistanceLabeledGraphNode vWithDistance = asDLGN(v);
			int alt = Numbers.add(this.current.tentativeDistance, this.map.getDistance(this.current, vWithDistance));
			// Relax (u,v,a)?
			if (v.relableAs(vWithDistance, alt))
				this.pathTo.put(vWithDistance, this.current);
		}

		// remove u from Q
		this.unvisited.safeRemove(this.current);
		
		if (this.unvisited.isEmpty()) {
			savePath(this.path);
		} else {
			// top part of the loop:
			// u := vertex in Q with smallest dist[]
			int minDist = Numbers.INFINITY;
			DistanceLabeledGraphNode u = null;
			for(DistanceLabeledGraphNode candidate : this.unvisited)
				if (candidate.tentativeDistance < minDist) {
					minDist = candidate.tentativeDistance;
					u = candidate;
				}

			rebind(u, u);
			return execute();
		}
		return this.destination;
	}

	void savePath(List<data.Node> path) {
		data.Node node = this.destination;
		do {
			path.add(0, node);
			node = this.pathTo.safeGet(node);
		} while (node != null);
	}

	/** print the nodes back to front starting with 'node' (like in the Ruby example). */
	public void print(Node as DistanceLabeledGraphNode node, boolean verbose) {
		DistanceLabeledGraphNode previous = null;
		while (node != null) {
			if (previous != null) {
				if (verbose)
					System.out.print(" - "+this.map.getDistance(node, previous)+" -> ");
				else
					System.out.print(" -> ");
			}
			System.out.print(node.getName());
			previous = node;
			node = asDLGN(this.pathTo.safeGet(node));
		}
		System.out.println();
	}

	/** Answer the given node's distance from the origin. */
	public int getDistanceFromOrigin(Node as DistanceLabeledGraphNode node) {
		return node.tentativeDistance;
	}
}
